"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.cleanChannelTags = exports.cleanQueryPolicy = exports.cleanEventPolicy = exports.assertQueryPolicy = exports.assertEventPolicy = exports.Channel = void 0;
const _ = require("lodash");
const utils_1 = require("./utils");
class Channel {
    constructor(pico, id, conf, familyChannelPicoID) {
        /**
         * A way to categorize channels i.e. bulk update of a group of channels
         */
        this.tags = [];
        /**
         * pico-engine had shared policies, however there were more problems by sharing that space
         * pico-framework policies are denormalized i.e. each channel owns their own copy
         */
        this.eventPolicy = {
            allow: [{ domain: "*", name: "*" }],
            deny: []
        };
        this.queryPolicy = {
            allow: [{ rid: "*", name: "*" }],
            deny: []
        };
        this.pico = pico;
        this.id = id;
        if (familyChannelPicoID) {
            this.tags = (conf && conf.tags) || [];
            this.familyChannelPicoID = familyChannelPicoID;
        }
        else {
            if (conf && conf.tags) {
                this.tags = cleanChannelTags(conf.tags);
            }
            // if not a family channel, use policies
            if (conf && conf.eventPolicy) {
                this.eventPolicy = cleanEventPolicy(conf.eventPolicy);
            }
            if (conf && conf.queryPolicy) {
                this.queryPolicy = cleanQueryPolicy(conf.queryPolicy);
            }
        }
    }
    update(conf) {
        if (this.familyChannelPicoID) {
            // can't change anything
            return;
        }
        if (conf.tags) {
            this.tags = cleanChannelTags(conf.tags);
        }
        if (conf.eventPolicy) {
            this.eventPolicy = cleanEventPolicy(conf.eventPolicy);
        }
        if (conf.queryPolicy) {
            this.queryPolicy = cleanQueryPolicy(conf.queryPolicy);
        }
        // NOTE: can't change familyChannelPicoID
    }
    assertEventPolicy(event, fromPicoID) {
        this.assertIfFamilyChannel(fromPicoID);
        assertEventPolicy(this.eventPolicy, event);
    }
    assertQueryPolicy(query, fromPicoID) {
        this.assertIfFamilyChannel(fromPicoID);
        assertQueryPolicy(this.queryPolicy, query);
    }
    assertIfFamilyChannel(fromPicoID) {
        if (this.familyChannelPicoID) {
            if (this.familyChannelPicoID === fromPicoID) {
                return; // good to go
            }
            else {
                throw new Error("This is a family channel and only the owner can use it.");
            }
        }
    }
    toReadOnly() {
        return Object.freeze({
            id: this.id,
            tags: this.tags.slice(0),
            eventPolicy: _.cloneDeep(this.eventPolicy),
            queryPolicy: _.cloneDeep(this.queryPolicy),
            familyChannelPicoID: this.familyChannelPicoID || null
        });
    }
    toDbPut() {
        return {
            type: "put",
            key: ["pico-channel", this.id],
            value: {
                id: this.id,
                picoId: this.pico.id,
                tags: this.tags,
                eventPolicy: this.eventPolicy,
                queryPolicy: this.queryPolicy,
                familyChannelPicoID: this.familyChannelPicoID || null
            }
        };
    }
    toDbDel() {
        return { type: "del", key: ["pico-channel", this.id] };
    }
    static fromDb(pico, val) {
        const chann = new Channel(pico, val.id, {
            tags: val.tags,
            eventPolicy: val.eventPolicy,
            queryPolicy: val.queryPolicy
        }, val.familyChannelPicoID);
        return chann;
    }
}
exports.Channel = Channel;
function assertEventPolicy(eventPolicy, event) {
    for (const p of eventPolicy.deny) {
        if (p.domain === "*" || p.domain === event.domain) {
            if (p.name === "*" || p.name === event.name) {
                throw new Error("Denied by channel policy");
            }
        }
    }
    let isAllowed = false;
    for (const p of eventPolicy.allow) {
        if (p.domain === "*" || p.domain === event.domain) {
            if (p.name === "*" || p.name === event.name) {
                isAllowed = true;
                break;
            }
        }
    }
    if (!isAllowed) {
        throw new Error("Not allowed by channel policy");
    }
}
exports.assertEventPolicy = assertEventPolicy;
function assertQueryPolicy(queryPolicy, query) {
    for (const p of queryPolicy.deny) {
        if (p.rid === "*" || p.rid === query.rid) {
            if (p.name === "*" || p.name === query.name) {
                throw new Error("Denied by channel policy");
            }
        }
    }
    let isAllowed = false;
    for (const p of queryPolicy.allow) {
        if (p.rid === "*" || p.rid === query.rid) {
            if (p.name === "*" || p.name === query.name) {
                isAllowed = true;
                break;
            }
        }
    }
    if (!isAllowed) {
        throw new Error("Not allowed by channel policy");
    }
}
exports.assertQueryPolicy = assertQueryPolicy;
function cleanEventPolicy(orig) {
    if (!orig ||
        Object.keys(orig)
            .sort()
            .join(",") !== "allow,deny" ||
        !Array.isArray(orig.allow) ||
        !Array.isArray(orig.deny)) {
        throw new TypeError(`EventPolicy expects {allow: EventPolicyRule[], deny: EventPolicyRule[]}`);
    }
    return {
        allow: orig.allow.map(cleanEventPolicyRule),
        deny: orig.deny.map(cleanEventPolicyRule)
    };
}
exports.cleanEventPolicy = cleanEventPolicy;
function cleanEventPolicyRule(orig) {
    if (!orig ||
        Object.keys(orig)
            .sort()
            .join(",") !== "domain,name" ||
        (0, utils_1.isNotStringOrBlank)(orig.domain) ||
        (0, utils_1.isNotStringOrBlank)(orig.name)) {
        throw new TypeError(`EventPolicyRule expects {domain: string, name: string}`);
    }
    return {
        domain: orig.domain.trim(),
        name: orig.name.trim()
    };
}
function cleanQueryPolicy(orig) {
    if (!orig ||
        Object.keys(orig)
            .sort()
            .join(",") !== "allow,deny" ||
        !Array.isArray(orig.allow) ||
        !Array.isArray(orig.deny)) {
        throw new TypeError(`QueryPolicy expects {allow: QueryPolicyRule[], deny: QueryPolicyRule[]}`);
    }
    return {
        allow: orig.allow.map(cleanQueryPolicyRule),
        deny: orig.deny.map(cleanQueryPolicyRule)
    };
}
exports.cleanQueryPolicy = cleanQueryPolicy;
function cleanQueryPolicyRule(orig) {
    if (!orig ||
        Object.keys(orig)
            .sort()
            .join(",") !== "name,rid" ||
        (0, utils_1.isNotStringOrBlank)(orig.name) ||
        (0, utils_1.isNotStringOrBlank)(orig.rid)) {
        throw new TypeError(`QueryPolicyRule expects {rid: string, name: string}`);
    }
    return {
        rid: orig.rid.trim(),
        name: orig.name.trim()
    };
}
function cleanChannelTags(tags) {
    if (!Array.isArray(tags)) {
        throw new TypeError("Channel `tags` must be an array of non-empty strings.");
    }
    return tags.map(tag => {
        if (typeof tag !== "string") {
            throw new TypeError("Channel `tags` must be an array of non-empty strings.");
        }
        tag = tag
            .replace(/\s+/g, " ")
            .trim()
            .replace(/[^a-z0-9_-]+/gi, "-")
            .replace(/-+/g, "-")
            .replace(/^-|-$/g, "")
            .toLowerCase();
        if (tag.length === 0) {
            throw new TypeError("Channel tag cannot be a blank string.");
        }
        if (tag === "system") {
            throw new TypeError('Cannot tag channel as "system".');
        }
        return tag;
    });
}
exports.cleanChannelTags = cleanChannelTags;
//# sourceMappingURL=Channel.js.map