import { Channel, ChannelConfig, ChannelReadOnly } from "./Channel";
import { PicoEvent, PicoEventPayload } from "./PicoEvent";
import { PicoFramework } from "./PicoFramework";
import { PicoQuery } from "./PicoQuery";
import { PicoTxn } from "./PicoQueue";
import { Ruleset, RulesetConfig, RulesetInstance } from "./Ruleset";
import { LevelBatch } from "./utils";
export interface PicoReadOnly {
    /**
     * The pico's id, used to correlate framework events.
     * This is also an ECI only the pico can use to talk to itself i.e. scheduled events
     */
    id: string;
    /**
     * The pico's parent ECI
     * This is null for the root pico.
     */
    parent: string | null;
    /**
     * List of ECIs to the pico's children
     */
    children: string[];
    /**
     * List of the pico's channels
     */
    channels: ChannelReadOnly[];
    /**
     * Rulesets installed on the pico
     */
    rulesets: PicoRulesetReadOnly[];
}
export interface PicoRulesetReadOnly {
    rid: string;
    config: RulesetConfig;
}
export interface NewPicoRuleset {
    rs: Ruleset;
    config?: RulesetConfig;
}
export interface NewPicoConfig {
    rulesets?: NewPicoRuleset[];
}
export declare class Pico {
    private pf;
    id: string;
    /**
     * The ECI string to the parent pico. Null if it's the root pico.
     */
    parent: string | null;
    /**
     * The ECI strings to the child pico's
     */
    children: string[];
    channels: {
        [eci: string]: Channel;
    };
    rulesets: {
        [rid: string]: {
            config: RulesetConfig;
            instance: RulesetInstance | null;
            startupError?: any;
        };
    };
    private queue;
    constructor(pf: PicoFramework, id: string);
    event(event: PicoEvent): Promise<string>;
    eventWait(event: PicoEvent): Promise<{
        eid: string;
        responses: any[];
    }>;
    eventQuery(event: PicoEvent, query: PicoQuery): Promise<any>;
    query(query: PicoQuery): Promise<any>;
    getCurrentTxn(): PicoTxn | undefined;
    waitFor(txnId: string): Promise<any>;
    newPico(conf?: NewPicoConfig): Promise<string>;
    delPico(eci: string): Promise<void>;
    /**
     * Recursively get the delete operations and ids
     * DO NOT mutate state, only build the operations
     */
    private delPicoDbOps;
    toReadOnly(): PicoReadOnly;
    toDbPut(): LevelBatch;
    toDbDel(): LevelBatch;
    static fromDb(pf: PicoFramework, val: any): Pico;
    newChannel(conf?: ChannelConfig, familyChannelPicoID?: string): Promise<Channel>;
    private newChannelBase;
    putChannel(eci: string, conf: ChannelConfig): Promise<Channel>;
    delChannel(eci: string): Promise<void>;
    install(rs: Ruleset, config?: RulesetConfig): Promise<void>;
    reInitRuleset(rs: Ruleset): Promise<void>;
    private installBase;
    uninstall(rid: string): Promise<void>;
    private uninstallBase;
    private assertInstalled;
    getEnt(rid: string, name: string): Promise<any>;
    putEnt(rid: string, name: string, value: any): Promise<void>;
    delEnt(rid: string, name: string): Promise<void>;
    private schedule;
    private current;
    private currentTxn;
    raiseEvent(domain: string, name: string, attrs: PicoEventPayload["attrs"], forRid?: string): void;
    clearSchedule(): void;
    addToSchedule(rid: string, event: PicoEvent): void;
    private addEventToSchedule;
    private doTxn;
}
