"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Pico = void 0;
const _ = require("lodash");
const Channel_1 = require("./Channel");
const dbRange_1 = require("./dbRange");
const PicoEvent_1 = require("./PicoEvent");
const PicoQueue_1 = require("./PicoQueue");
const RulesetContext_1 = require("./RulesetContext");
class Pico {
    constructor(pf, id) {
        this.pf = pf;
        /**
         * The ECI string to the parent pico. Null if it's the root pico.
         */
        this.parent = null;
        /**
         * The ECI strings to the child pico's
         */
        this.children = [];
        this.channels = {};
        this.rulesets = {};
        this.schedule = [];
        this.id = id;
        this.channels[id] = new Channel_1.Channel(this, id, { tags: ["system", "self"] }, id);
        this.queue = new PicoQueue_1.PicoQueue(this.id, this.doTxn.bind(this), (ev) => this.pf.emit(ev));
    }
    async event(event) {
        const eid = this.pf.genID();
        this.queue.push({
            id: eid,
            kind: "event",
            event,
        });
        return eid;
    }
    eventWait(event) {
        const eid = this.pf.genID();
        this.queue.push({
            id: eid,
            kind: "event",
            event,
        });
        return this.queue.waitFor(eid);
    }
    async eventQuery(event, query) {
        const eid = this.pf.genID();
        this.queue.push({
            id: eid,
            kind: "event",
            event,
        });
        const eidQ = eid + ".q";
        this.queue.push({
            id: eidQ,
            kind: "query",
            query,
        });
        const p0 = this.queue.waitFor(eid);
        const p1 = this.queue.waitFor(eidQ);
        await p0; // this may throwup
        return p1;
    }
    async query(query) {
        const eid = this.pf.genID();
        this.queue.push({
            id: eid,
            kind: "query",
            query,
        });
        return this.queue.waitFor(eid);
    }
    getCurrentTxn() {
        return this.queue.getCurrentTxn();
    }
    waitFor(txnId) {
        return this.queue.waitFor(txnId);
    }
    async newPico(conf) {
        const child = new Pico(this.pf, this.pf.genID());
        const parentChannel = this.newChannelBase({ tags: ["system", "parent"] }, child.id);
        child.parent = parentChannel.id;
        const childChannel = child.newChannelBase({ tags: ["system", "child"] }, this.id);
        child.channels[childChannel.id] = childChannel;
        this.children.push(childChannel.id);
        const dbOps = [
            this.toDbPut(),
            child.toDbPut(),
            parentChannel.toDbPut(),
            childChannel.toDbPut(),
        ];
        try {
            await this.pf.db.batch(dbOps);
        }
        catch (err) {
            this.children = this.children.filter((c) => c !== childChannel.id);
            throw err;
        }
        this.channels[parentChannel.id] = parentChannel;
        this.pf.addPico(child);
        if (conf && conf.rulesets) {
            for (const rs of conf.rulesets) {
                try {
                    await child.install(rs.rs, rs.config);
                }
                catch (err) {
                    this.pf.emit({
                        type: "newPicoInstallError",
                        picoId: child.id,
                        rid: rs.rs.rid,
                        config: rs.config,
                        error: err,
                    });
                }
            }
        }
        return childChannel.id;
    }
    async delPico(eci) {
        if (this.children.indexOf(eci) < 0) {
            throw new Error(`delPico(${eci}) - not found in children ECIs`);
        }
        const { pico } = this.pf.lookupChannel(eci);
        const { ops, picoIds } = await pico.delPicoDbOps();
        if (pico.parent) {
            ops.push(this.channels[pico.parent].toDbDel());
        }
        this.children = this.children.filter((c) => c !== eci);
        ops.push(this.toDbPut());
        try {
            await this.pf.db.batch(ops);
        }
        catch (err) {
            this.children.push(eci); // restore state
            throw err;
        }
        if (pico.parent) {
            delete this.channels[pico.parent];
        }
        for (const id of picoIds) {
            this.pf.removePico(id);
        }
    }
    /**
     * Recursively get the delete operations and ids
     * DO NOT mutate state, only build the operations
     */
    async delPicoDbOps() {
        let ops = [];
        let picoIds = [];
        for (const eci of this.children) {
            // recursive delete
            const { pico } = this.pf.lookupChannel(eci);
            const sub = await pico.delPicoDbOps();
            ops = ops.concat(sub.ops);
            picoIds = picoIds.concat(sub.picoIds);
        }
        for (const channel of Object.values(this.channels)) {
            ops.push(channel.toDbDel());
        }
        for (const rid of Object.keys(this.rulesets)) {
            ops = ops.concat(await this.uninstallBase(rid));
        }
        ops.push(this.toDbDel());
        picoIds.push(this.id);
        return { ops, picoIds };
    }
    toReadOnly() {
        const data = {
            id: this.id,
            parent: this.parent,
            children: this.children.slice(0),
            channels: Object.values(this.channels).map((c) => c.toReadOnly()),
            rulesets: Object.keys(this.rulesets).map((rid) => ({
                rid,
                config: this.rulesets[rid].config,
            })),
        };
        data.channels = _.sortBy(data.channels, (c) => {
            return c.id;
        });
        return Object.freeze(data);
    }
    toDbPut() {
        return {
            type: "put",
            key: ["pico", this.id],
            value: {
                id: this.id,
                parent: this.parent,
                children: this.children.slice(0),
            },
        };
    }
    toDbDel() {
        return { type: "del", key: ["pico", this.id] };
    }
    static fromDb(pf, val) {
        const pico = new Pico(pf, val.id);
        pico.parent = val.parent;
        pico.children = val.children;
        return pico;
    }
    async newChannel(conf, familyChannelPicoID) {
        const chann = this.newChannelBase(conf, familyChannelPicoID);
        await this.pf.db.batch([chann.toDbPut()]);
        this.channels[chann.id] = chann;
        return chann;
    }
    newChannelBase(conf, familyChannelPicoID) {
        const chann = new Channel_1.Channel(this, this.pf.genID(), conf, familyChannelPicoID);
        return chann;
    }
    async putChannel(eci, conf) {
        const chann = this.channels[eci];
        if (!chann) {
            throw new Error(`ECI not found ${eci}`);
        }
        if (chann.familyChannelPicoID == this.id) {
            throw new Error("Cannot edit the self channel.");
        }
        chann.update(conf);
        await this.pf.db.batch([chann.toDbPut()]);
        return chann;
    }
    async delChannel(eci) {
        const chann = this.channels[eci];
        if (!chann) {
            throw new Error(`ECI not found ${eci}`);
        }
        if (chann.familyChannelPicoID) {
            throw new Error("Cannot delete family channels.");
        }
        await this.pf.db.del(["pico-channel", chann.id]);
        delete this.channels[eci];
    }
    async install(rs, config = {}) {
        const { instance, dbPut } = await this.installBase(rs, config);
        await this.pf.db.batch([dbPut]);
        this.rulesets[rs.rid] = { config, instance };
    }
    async reInitRuleset(rs) {
        const rid = rs.rid;
        if (this.rulesets[rid]) {
            const config = this.rulesets[rid].config;
            const ctx = (0, RulesetContext_1.createRulesetContext)(this.pf, this, { rid, config });
            const instance = await rs.init(ctx, this.pf.environment);
            this.rulesets[rid].instance = instance;
            delete this.rulesets[rid].startupError;
        }
    }
    async installBase(rs, config = {}) {
        // even if we already have that rid installed, we need to init again
        // b/c the configuration may have changed
        const ctx = (0, RulesetContext_1.createRulesetContext)(this.pf, this, {
            rid: rs.rid,
            config,
        });
        const instance = await rs.init(ctx, this.pf.environment);
        return {
            instance,
            dbPut: {
                type: "put",
                key: ["pico-ruleset", this.id, rs.rid],
                value: {
                    rid: rs.rid,
                    config,
                },
            },
        };
    }
    async uninstall(rid) {
        const ops = await this.uninstallBase(rid);
        await this.pf.db.batch(ops);
        delete this.rulesets[rid];
    }
    async uninstallBase(rid) {
        const ops = await (0, dbRange_1.dbRange)(this.pf.db, { prefix: ["entvar", this.id, rid] }, (data) => {
            const delEnt = {
                type: "del",
                key: data.key,
            };
            return delEnt;
        });
        ops.push({
            type: "del",
            key: ["pico-ruleset", this.id, rid],
        });
        return ops;
    }
    assertInstalled(rid) {
        if (!this.rulesets[rid]) {
            throw new Error(`Not installed ${rid}`);
        }
    }
    async getEnt(rid, name) {
        this.assertInstalled(rid);
        let data;
        try {
            data = await this.pf.db.get(["entvar", this.id, rid, name]);
        }
        catch (err) {
            if (err.notFound) {
                return null;
            }
        }
        return data;
    }
    async putEnt(rid, name, value) {
        this.assertInstalled(rid);
        const key = ["entvar", this.id, rid, name];
        if (value === null || value === undefined) {
            await this.pf.db.del(key);
        }
        else {
            await this.pf.db.put(key, value);
        }
    }
    async delEnt(rid, name) {
        this.assertInstalled(rid);
        await this.pf.db.del(["entvar", this.id, rid, name]);
    }
    raiseEvent(domain, name, attrs, forRid) {
        const event = (0, PicoEvent_1.cleanEvent)({
            eci: (this.current && this.current.event.eci) || "[raise]",
            domain,
            name,
            data: { attrs },
        });
        if (typeof forRid === "string") {
            this.addToSchedule(forRid, event);
        }
        else {
            this.addEventToSchedule(event);
        }
    }
    clearSchedule() {
        this.schedule = [];
        if (this.currentTxn && this.currentTxn.kind === "event") {
            this.pf.emit({
                type: "eventScheduleCleared",
                picoId: this.id,
                txn: this.currentTxn,
            });
        }
    }
    addToSchedule(rid, event) {
        this.schedule.push({ rid, event });
        if (this.currentTxn && this.currentTxn.kind === "event") {
            this.pf.emit({
                type: "eventScheduleAdded",
                picoId: this.id,
                txn: this.currentTxn,
                rid,
                event,
            });
        }
    }
    addEventToSchedule(event) {
        for (const rid of Object.keys(this.rulesets)) {
            this.addToSchedule(rid, event);
        }
    }
    async doTxn(txn) {
        var _a, _b;
        this.currentTxn = txn;
        try {
            switch (txn.kind) {
                case "event":
                    this.schedule = []; // reset schedule every new event
                    this.addEventToSchedule(txn.event);
                    const eid = txn.id;
                    const responses = [];
                    try {
                        while ((this.current = this.schedule.shift())) {
                            const rs = this.rulesets[this.current.rid];
                            if ((_a = rs === null || rs === void 0 ? void 0 : rs.instance) === null || _a === void 0 ? void 0 : _a.event) {
                                // must process one event at a time to maintain the pico's single-threaded guarantee
                                const response = await rs.instance.event(this.current.event, eid);
                                responses.push(response);
                            }
                        }
                    }
                    finally {
                        this.current = undefined;
                    }
                    return { eid, responses };
                case "query":
                    const rs = this.rulesets[txn.query.rid];
                    if (!rs) {
                        throw new Error(`Pico doesn't have ${txn.query.rid} installed.`);
                    }
                    const qfn = ((_b = rs === null || rs === void 0 ? void 0 : rs.instance) === null || _b === void 0 ? void 0 : _b.query) && rs.instance.query[txn.query.name];
                    if (!qfn) {
                        throw new Error(`Ruleset ${txn.query.rid} does not have query function "${txn.query.name}"`);
                    }
                    const data = await qfn(txn.query, txn.id);
                    return data;
            }
        }
        finally {
            this.currentTxn = undefined;
        }
    }
}
exports.Pico = Pico;
//# sourceMappingURL=Pico.js.map