import { PicoDb } from "./utils";
import { Channel } from "./Channel";
import { Pico } from "./Pico";
import { PicoEvent } from "./PicoEvent";
import { PicoFrameworkEvent } from "./PicoFrameworkEvent";
import { PicoQuery } from "./PicoQuery";
import { Ruleset, RulesetConfig } from "./Ruleset";
export type RulesetLoader = (picoId: string, rid: string, config: RulesetConfig) => Ruleset | Promise<Ruleset>;
type OnFrameworkEvent = (event: PicoFrameworkEvent) => void;
export interface PicoFrameworkConf {
    rulesetLoader: RulesetLoader;
    /**
     * Specify how data should be persisted.
     */
    db: PicoDb;
    /**
     * Function that is called on a pico framework event
     */
    onFrameworkEvent?: OnFrameworkEvent;
    /**
     * Set true if you want to trust the timestamp on events entering the system.
     */
    useEventInputTime?: boolean;
    /**
     * Pass optional data to each ruleset init function
     */
    environment?: any;
    /**
     * Optionally override the function that generates random-unique ids.
     *
     * This is mostly intended for testing purposes.
     */
    genID?: () => string;
}
export declare class PicoFramework {
    db: PicoDb;
    private rootPico_?;
    get rootPico(): Pico;
    private picos;
    private startupP;
    genID: () => string;
    private rulesetLoader;
    readonly environment?: any;
    private useEventInputTime;
    /**
     * not using EventEmitter b/c I want it typed checked and limited.
     */
    private onFrameworkEvent?;
    constructor(conf: PicoFrameworkConf);
    private startup;
    start(): Promise<void>;
    cleanEvent(event: PicoEvent): PicoEvent;
    /**
     * Signal an event and return the txn id
     *
     * @param event
     * @param fromPicoId This is for detecting family relationships
     */
    event(event: PicoEvent, fromPicoId?: string): Promise<string>;
    /**
     * Same as `event` but will wait until it finishes processing
     *
     * @param event
     * @param fromPicoId
     */
    eventWait(event: PicoEvent, fromPicoId?: string): Promise<{
        eid: string;
        responses: any[];
    }>;
    /**
     * Same as `event` but will immediately run a query after the event and return the query result
     *
     * @param event
     * @param query
     * @param fromPicoId
     */
    eventQuery(event: PicoEvent, query: PicoQuery, fromPicoId?: string): Promise<any>;
    query(query: PicoQuery, fromPicoId?: string): Promise<any>;
    lookupChannel(eci: string): Channel;
    getPico(eci: string): Pico;
    reInitRuleset(rs: Ruleset): void;
    /**
     * @ignore
     *
     * NOTE: not using EventEmitter so we can have type information and it can be synchronous
     */
    emit(event: PicoFrameworkEvent): void;
    /**
     * @ignore
     */
    addPico(pico: Pico): void;
    /**
     * @ignore
     */
    removePico(picoId: string): void;
    /**
     * Return the number of pico's
     * Useful for testing
     */
    numberOfPicos(): number;
}
export {};
