"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PicoFramework = void 0;
const cuid = require("cuid");
const Channel_1 = require("./Channel");
const dbRange_1 = require("./dbRange");
const Pico_1 = require("./Pico");
const PicoEvent_1 = require("./PicoEvent");
const PicoQuery_1 = require("./PicoQuery");
class PicoFramework {
    get rootPico() {
        if (!this.rootPico_) {
            throw new Error("rootPico has not started up yet");
        }
        return this.rootPico_;
    }
    constructor(conf) {
        this.picos = [];
        this.useEventInputTime = false;
        this.db = conf.db;
        this.rulesetLoader = conf && conf.rulesetLoader;
        this.genID = (conf && conf.genID) || cuid;
        this.environment = conf && conf.environment;
        this.onFrameworkEvent = conf && conf.onFrameworkEvent;
        this.useEventInputTime = !!(conf && conf.useEventInputTime);
        this.startupP = this.startup();
    }
    async startup() {
        this.emit({ type: "startup" });
        await (0, dbRange_1.dbRange)(this.db, { prefix: ["pico"] }, (data) => {
            const pico = Pico_1.Pico.fromDb(this, data.value);
            this.picos.push(pico);
        });
        await (0, dbRange_1.dbRange)(this.db, { prefix: ["pico-channel"] }, (data) => {
            const { picoId } = data.value;
            const pico = this.picos.find((pico) => pico.id === picoId);
            if (!pico) {
                throw new Error(`Missing picoId ${picoId}`);
            }
            const chann = Channel_1.Channel.fromDb(pico, data.value);
            pico.channels[chann.id] = chann;
        });
        // get rids that are used, and load them
        // install separate so they are all loaded/warmed up first in case of dependencies
        const toInstall = [];
        await (0, dbRange_1.dbRange)(this.db, { prefix: ["pico-ruleset"] }, async (data) => {
            const picoId = data.key[1];
            const rid = data.key[2];
            const config = data.value.config;
            const pico = this.picos.find((pico) => pico.id === picoId);
            if (!pico) {
                throw new Error(`Missing picoId ${picoId}`);
            }
            // load ruleset map so when rulesets startup they can see all the available ruleset+config on the pico
            pico.rulesets[rid] = { config, instance: null };
            try {
                const rs = await this.rulesetLoader(picoId, rid, data.value.config);
                toInstall.push({ pico, rs, config: data.value.config });
            }
            catch (error) {
                pico.rulesets[rid] = { config, instance: null, startupError: error };
                this.emit({
                    type: "startupRulesetInitError",
                    picoId,
                    rid,
                    config: data.value.config,
                    error,
                });
            }
        });
        for (const { pico, rs, config } of toInstall) {
            try {
                await pico.install(rs, config);
            }
            catch (error) {
                this.emit({
                    type: "startupRulesetInitError",
                    picoId: pico.id,
                    rid: rs.rid,
                    config,
                    error,
                });
            }
        }
        let rootId;
        try {
            rootId = await this.db.get(["root-pico"]);
        }
        catch (err) {
            if (err.notFound) {
                rootId = null;
            }
            else {
                throw err;
            }
        }
        if (rootId) {
            this.rootPico_ = this.picos.find((pico) => pico.id === rootId);
            if (!this.rootPico_) {
                throw new Error(`Bad root pico ID ${rootId}`);
            }
        }
        else {
            const pico = new Pico_1.Pico(this, this.genID());
            await this.db.batch([
                pico.toDbPut(),
                { type: "put", key: ["root-pico"], value: pico.id },
            ]);
            this.rootPico_ = pico;
            this.picos.push(pico);
        }
        this.emit({ type: "startupDone" });
    }
    start() {
        return this.startupP;
    }
    cleanEvent(event) {
        if (this.useEventInputTime) {
            return (0, PicoEvent_1.cleanEvent)(event, event.time);
        }
        return (0, PicoEvent_1.cleanEvent)(event);
    }
    /**
     * Signal an event and return the txn id
     *
     * @param event
     * @param fromPicoId This is for detecting family relationships
     */
    async event(event, fromPicoId) {
        event = this.cleanEvent(event);
        const channel = this.lookupChannel(event.eci);
        channel.assertEventPolicy(event, fromPicoId);
        return channel.pico.event(event);
    }
    /**
     * Same as `event` but will wait until it finishes processing
     *
     * @param event
     * @param fromPicoId
     */
    async eventWait(event, fromPicoId) {
        event = this.cleanEvent(event);
        const channel = this.lookupChannel(event.eci);
        channel.assertEventPolicy(event, fromPicoId);
        return channel.pico.eventWait(event);
    }
    /**
     * Same as `event` but will immediately run a query after the event and return the query result
     *
     * @param event
     * @param query
     * @param fromPicoId
     */
    async eventQuery(event, query, fromPicoId) {
        event = this.cleanEvent(event);
        query = (0, PicoQuery_1.cleanQuery)(query);
        if (query.eci !== event.eci) {
            throw new Error("eventQuery must use the same channel");
        }
        const channel = this.lookupChannel(event.eci);
        channel.assertEventPolicy(event, fromPicoId);
        channel.assertQueryPolicy(query, fromPicoId);
        return channel.pico.eventQuery(event, query);
    }
    async query(query, fromPicoId) {
        query = (0, PicoQuery_1.cleanQuery)(query);
        const channel = this.lookupChannel(query.eci);
        channel.assertQueryPolicy(query, fromPicoId);
        return channel.pico.query(query);
    }
    lookupChannel(eci) {
        for (const pico of this.picos) {
            if (pico.channels[eci]) {
                return pico.channels[eci];
            }
        }
        throw new Error(`ECI not found ${eci}`);
    }
    getPico(eci) {
        for (const pico of this.picos) {
            if (pico.channels[eci]) {
                return pico;
            }
        }
        throw new Error(`ECI not found ${eci}`);
    }
    reInitRuleset(rs) {
        for (const pico of this.picos) {
            pico.reInitRuleset(rs).catch((err) => {
                var _a;
                this.emit({
                    type: "reInitRulesetError",
                    picoId: pico.id,
                    rid: rs.rid,
                    config: (_a = pico.rulesets[rs.rid]) === null || _a === void 0 ? void 0 : _a.config,
                    error: err,
                });
            });
        }
    }
    /**
     * @ignore
     *
     * NOTE: not using EventEmitter so we can have type information and it can be synchronous
     */
    emit(event) {
        if (!this.onFrameworkEvent)
            return;
        this.onFrameworkEvent(event);
    }
    /**
     * @ignore
     */
    addPico(pico) {
        this.picos.push(pico);
    }
    /**
     * @ignore
     */
    removePico(picoId) {
        this.picos = this.picos.filter((p) => p.id !== picoId);
        this.emit({ type: "picoDeleted", picoId });
    }
    /**
     * Return the number of pico's
     * Useful for testing
     */
    numberOfPicos() {
        return this.picos.length;
    }
}
exports.PicoFramework = PicoFramework;
//# sourceMappingURL=PicoFramework.js.map